#include <iomanip>
#include <iostream>
#include <vector>
#include <map>
#include <Random.h>
#include "FrameList.h"
#include "errstr.h"
#include "Xmm.h"

#include <ctime>

Frames::Frames(double start, double stop, double fracExp) :
  Gti(start,stop) , _fracExp(fracExp)
{

}

double Frames::getFracExp()
{
  return _fracExp;
}

void FrameList::printFrames()
{
	for (unsigned int n = 0 ; n < (*this).size() ; n++)
		errstr << setprecision(19) << "FrameBin= [ "<< (*this)[n].start() << " , "
			<< (*this)[n].stop() << " ]   FRACEXP= "<< (*this)[n].getFracExp()
			<< Sas::message(Msg::Lib, Msg::Verbose);
}

void FrameList::printDeadTime()
{
	for (unsigned int n = 0 ; n < (*this).size()-1 ; n++)
		errstr << "DeadTimeBin[ "<< n << "]= "<<
			(*this)[n+1].stop()-(*this)[n].stop()
			<< Sas::message(Msg::Lib, Msg::Verbose);
}

Exposure::Exposure(FrameList frameList) : _frameList(frameList)
{

}

/*
 * Sum of all the fracexp of all the frames inside
 * each light curve gti.
 */
void Exposure::doCorr(LCGrid *lc, std::multimap<double,int>* events, std::string inst, bool rndFlag, std::string subMode)
{

	std::vector<double> frameTime, gtiTime;
	// Initialize _fracExp vector
	  for (unsigned int i = 0 ; i<=lc->_cells ; i++)
	  {
		  _expoCorr.push_back(0.0);
		  frameTime.push_back(0.0);
		  gtiTime.push_back(0.0);
	  }


	unsigned int i = 0;  // GTI counter
	unsigned int n = 0;  // Frames counter


	// Skip all the frames that starts earlier than the first GTI
	while ((_frameList[n].start() < lc->_gtiList[0].start()) &&
			( _frameList[n].stop() < lc->_gtiList[0].start() ) )
	{
		errstr<< "Frame[ " << n << " ] = [ "<< _frameList[n].start() << " , "
				<< _frameList[n].stop() << " ]  SKIPPED"<< Sas::message(Msg::Lib, Msg::Verbose);
		n++;
	}

	while ( ( i != lc->_gtiList.size() ) && (n != _frameList.size() ) )
	{
		errstr<< setprecision(19)<< "GTI[ "<<i<<" ] = [ "<< lc->_gtiList[i].start() << " , "
			<< lc->_gtiList[i].stop() << " ]"<< Sas::message(Msg::Lib, Msg::Verbose);
		//double fracExp = 0.0;
		double frameBinTime = 0.0;
		bool nextGTI = false;
		while ( (n != _frameList.size() )
				&& ( nextGTI == false) )
		{
			errstr<< "Frame[ " << n << " ] = [ "<< _frameList[n].start() << " , "
					<< _frameList[n].stop() << " ]"<< Sas::message(Msg::Lib, Msg::Verbose);

			if ( (_frameList[n].start() <= lc->_gtiList[i].start()) )
				if ( (_frameList[n].stop() > lc->_gtiList[i].start() )
						&&	(_frameList[n].stop() <= lc->_gtiList[i].stop() ) )
				{
					errstr << "First frame inside the bin[ "<< n << " ] = [ "<< _frameList[n].start() << " , "
					<< _frameList[n].stop() << " ]" << Sas::message(Msg::Lib, Msg::Verbose);
					//fracExp = ( (_frameList[n].stop()-lc->_gtiList[i].start())/
					//		(_frameList[n].stop() - _frameList[n].start()) )*_frameList[n]._fracExp;

					//frameBinTime = frameBinTime +fracExp*(_frameList[n].stop()-lc->_gtiList[i].start());
					frameBinTime = frameBinTime +_frameList[n]._fracExp*(_frameList[n].stop()-lc->_gtiList[i].start());
					// Randomize PN events.
					if ( (inst == "EPN") && (rndFlag == false) )
						if ( (Xmm::toMode(subMode) !=  Xmm::FAST_TIMING) && (Xmm::toMode(subMode) !=  Xmm::FAST_BURST )  )
							randomizePNEvents(lc, events, i,n);
						else
						{
							errstr << "Randomization for PN Timing or Burst not implemented"
							<< Sas::warning("PNTimingRand");
						}
				}
				else if (_frameList[n].stop() >= lc->_gtiList[i].stop() )
				{
					errstr << "Frame bigger then GTI bin[ "<< n << " ] = [ "<< _frameList[n].start() << " , "
					<< _frameList[n].stop() << " ]" << Sas::message(Msg::Lib, Msg::Verbose);
					//fracExp = ( (lc->_gtiList[i].stop()-lc->_gtiList[i].start())/
					//		(_frameList[n].stop() - _frameList[n].start()) )*_frameList[n]._fracExp;

					//frameBinTime = frameBinTime +fracExp*(lc->_gtiList[i].stop()-lc->_gtiList[i].start());
					frameBinTime = frameBinTime +_frameList[n]._fracExp*(lc->_gtiList[i].stop()-lc->_gtiList[i].start());
					nextGTI = true;
					n--;
				}

			if  ( ( _frameList[n].start() >= lc->_gtiList[i].start() )
					&& ( _frameList[n].start() < lc->_gtiList[i].stop() ) )
			{
				if (_frameList[n].stop() <= lc->_gtiList[i].stop() )
				{
					frameBinTime = frameBinTime +
						_frameList[n]._fracExp*(_frameList[n].stop()-_frameList[n].start());
				}
				else if ( _frameList[n].stop() > lc->_gtiList[i].stop())
				{
					errstr << "Last frame in current bin[ "<< n << " ] = [ " << _frameList[n].start() << " , "
						<< _frameList[n].stop() << " ]" << Sas::message(Msg::Lib, Msg::Verbose);
					//fracExp = ( ( lc->_gtiList[i].stop()-_frameList[n].start() ) /
					//		( _frameList[n].stop() - _frameList[n].start()))*_frameList[n]._fracExp;
					//frameBinTime = frameBinTime + fracExp*( lc->_gtiList[i].stop()-_frameList[n].start()  );
					frameBinTime = frameBinTime + _frameList[n]._fracExp*( lc->_gtiList[i].stop()-_frameList[n].start()  );

					// Randomize PN events.
					if ( (inst == "EPN") && (rndFlag == false) )
						if ( (Xmm::toMode(subMode) !=  Xmm::FAST_TIMING) && (Xmm::toMode(subMode) !=  Xmm::FAST_BURST )  )
							randomizePNEvents(lc, events, i,n);
						else
							{
								errstr << "Randomization for PN Timing or Burst not implemented"
								<< Sas::warning("PNTimingRand");
							}

					nextGTI = true;
					n--;
				}
			}
			else if  ( _frameList[n].start() >= lc->_gtiList[i].stop() )
			{
				errstr << "Frame in next bin. " << Sas::message(Msg::Lib, Msg::Verbose);
				nextGTI = true;
				n--;
			}

			n++;
		} // Frame while
		int bin = lc->cell(((lc->_gtiList[i].stop() + lc->_gtiList[i].start())/2.0));

		frameTime[bin] += frameBinTime;
		gtiTime[bin] += (lc->_gtiList[i].stop() - lc->_gtiList[i].start());
		i++;
	} // GTI while

	for (unsigned int i = 0 ; i < _expoCorr.size() ; i++ )
	{
		_expoCorr[i] +=   gtiTime[i] == 0.0 ? 0.0 : frameTime[i]/lc->getBinSize(i);
		lc->_fracExp[i] =  frameTime[i]/lc->getBinSize(i);
		lc->_effBinTime[i] = frameTime[i];
	}

}

void Exposure::randomizePNEvents(LCGrid *lc,  std::multimap<double,int>* events, unsigned int gtiBin, unsigned int frameBin)
{
	errstr <<"Randomizing PN events... "<< Sas::message(Msg::Lib, Msg::Verbose);
	typedef std::multimap<double,int>::const_iterator it;
	it itLow, itUp;

	itLow = events->lower_bound(_frameList[frameBin].start());
	itUp = events->upper_bound(_frameList[frameBin].stop());

	for (it iter = itLow ; iter != itUp ; iter++ )
	{
		if ((*iter).second ==  lc->_ccd )
		{
		int bin = lc->cell(((lc->_gtiList[gtiBin].stop() + lc->_gtiList[gtiBin].start())/2.0));
		double fracTime = (lc->_gtiList[gtiBin].stop()-_frameList[frameBin].start())/
		( _frameList[frameBin].stop()-_frameList[frameBin].start());

		if (fracTime >= 0.5)
			lc->_counts[bin] = lc->_counts[bin] - 1;
		else
			lc->_counts[bin+1] = lc->_counts[bin+1] - 1;

		if(randGen.rand() <= fracTime)
			lc->_counts[bin] = lc->_counts[bin] + 1;
		else
			lc->_counts[bin+1] = lc->_counts[bin+1] + 1;
		}
	}
}

void Exposure::printCorr()
{
	for (unsigned int i = 0 ; i < _expoCorr.size() ; i++ )
		errstr << "ExpoCorr["<<i<<"]= "<<  _expoCorr[i] << Sas::message(Msg::Lib, Msg::Sparse);
}


DeadTime::DeadTime(FrameList frameList) : _frameList(frameList)
{

}

void DeadTime::doCorr(LCGrid *lc)
{
	std::vector<double> deadTime, gtiTime;
	// Initialize _fracExp vector
	  for (unsigned int i = 0 ; i<=lc->_cells ; i++)
	  {
		  _deadTimeCorr.push_back(0.0);
		  deadTime.push_back(0.0);
		  gtiTime.push_back(0.0);
	  }

	unsigned int i = 0;  // GTI counter
	unsigned int n = 0;  // Frames counter

	// Skip all the frames that starts earlier than the first GTI
	while ((_frameList[n].start() < lc->_gtiList[0].start()) &&
			( _frameList[n].stop() <= lc->_gtiList[0].stop() ) )
		n++;

	while( (i != lc->_gtiList.size()) &&  (n != _frameList.size() ) )
	{
		errstr  << setprecision(19)<< "GTI[ "<<i<<"]= "<< lc->_gtiList[i].start() << " , "
			<< lc->_gtiList[i].stop() << " ]"<< Sas::message(Msg::Lib, Msg::Verbose);
		double deadBinTime = 0.0;
		bool nextGTI = false;
		while ( (n != _frameList.size() ) && ( nextGTI == false) )
		{
			errstr << "Frame =[ "<< _frameList[n].start() << " , "
				<< _frameList[n].stop() << " ]"<< Sas::message(Msg::Lib, Msg::Verbose);
			if ( (n == 0) && ( ( _frameList[n].start() > lc->_gtiList[i].start() )
							&& ( _frameList[n].start() <= lc->_gtiList[i].stop() )  ) )
			{
				errstr << "First Frame =[ "<< _frameList[n].start() << " , "
				<< _frameList[n].stop() << " ]" <<Sas::message(Msg::Lib, Msg::Verbose);
				deadBinTime = deadBinTime + (_frameList[n].start()-lc->_gtiList[i].start());

			}
			else if ( ( _frameList[n].start() > lc->_gtiList[i].start() )
					&& ( _frameList[n].start() <= lc->_gtiList[i].stop() ) )
			{
				if (  _frameList[n-1].stop() >= lc->_gtiList[i].start() )
				{
					deadBinTime = deadBinTime + ( _frameList[n].start()-_frameList[n-1].stop()  );

				}
				else
				{
					errstr << "First deadTime inside current bin =[ "<< _frameList[n].start() << " , "
						<< _frameList[n].stop() << " ]" <<Sas::message(Msg::Lib, Msg::Verbose);
					deadBinTime = deadBinTime + (_frameList[n].start()-lc->_gtiList[i].start());

				}
			}
			else if ( ( _frameList[n].start() >= lc->_gtiList[i].stop() )
					&& ( ( _frameList[n-1].stop() >= lc->_gtiList[i].start() )
							&& ( _frameList[n-1].stop() < lc->_gtiList[i].stop() ) ) )
			{
				errstr << "Last deadTime inside current bin =[" << _frameList[n].start() << " , "
				<< _frameList[n].stop() << " ]" << Sas::message(Msg::Lib, Msg::Verbose);
				deadBinTime = deadBinTime + ( lc->_gtiList[i].stop()-_frameList[n-1].stop()  );

			}

			if ( _frameList[n].stop() >= lc->_gtiList[i].stop() )
			{
				nextGTI = true;

			}

			n++;
		}

		int bin = lc->cell(((lc->_gtiList[i].stop() + lc->_gtiList[i].start())/2.0));
		deadTime[bin] += deadBinTime;
		gtiTime[bin] += (lc->_gtiList[i].stop() - lc->_gtiList[i].start());
		i++;

	} // GTI while

	for (unsigned int i = 0 ; i <= _deadTimeCorr.size() ; i++ )
	{
		_deadTimeCorr[i] +=   gtiTime[i] == 0.0 ? 0.0 : deadTime[i]/lc->getBinSize(i);
		lc->_fracExp[i] =  (lc->_effBinTime[i]-deadTime[i])/lc->getBinSize(i);
		lc->_effBinTime[i] = lc->_effBinTime[i]-deadTime[i];
	}

}

void DeadTime::printCorr()
{
	for (unsigned int i = 0 ; i < _deadTimeCorr.size() ; i++ )
		errstr << "DeadTimeCorr["<<i<<"]= "<< _deadTimeCorr[i] << Sas::message(Msg::Lib, Msg::Sparse);
}


